using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using QiHe.CodeLib;

namespace QiHe.RealMedia
{
    public partial class FieldDef
    {
        public object Decode(DataReader reader, Record record, FormatDef format, out bool hasError)
        {
            hasError = false;
            if (Type.StartsWith("Record"))
            {
                int size = GetSize(record);
                byte[] data = reader.ReadBytes(size);
                RecordDef recordDef = GetRecordDef(record, format);
                if (recordDef != null)
                {
                    Record subRecord = recordDef.Decode(new DataReader(new MemoryStream(data)), format);
                    subRecord.Data = data;
                    hasError = false;
                    return subRecord;
                }
                else
                {
                    return data;
                }
            }
            else if (Type.EndsWith("[]"))
            {
                string type = Type.TrimEnd('[', ']');
                int size = GetSize(record);
                if (type == "Byte")
                {
                    return reader.ReadBytes(size);
                }
                else if (type == "UInt2")
                {
                    return reader.ReadBytes(size * 2);
                }
                else if (type == "UInt4")
                {
                    return reader.ReadBytes(size * 4);
                }
                else
                {
                    RecordDef recordDef = format[type];
                    return new RecordCollection(format, recordDef, reader, size);
                }
            }
            else
            {
                string type = Type + (Size * 8) + Endian;
                switch (type)
                {
                    case "UInt8LE":
                    case "UInt8BE":
                        return reader.ReadByte();
                    case "UInt32LE":
                        return reader.ReadUInt32LE();
                    case "UInt16BE":
                        return reader.ReadUInt16BE();
                    case "UInt32BE":
                        return reader.ReadUInt32BE();
                    default:
                        int size = GetSize(record);
                        byte[] data = reader.ReadBytes(size);

                        switch (Type)
                        {
                            case "AsciiString":
                                return TextEncoding.SafeDecodeString(Encoding.ASCII, data);
                            case "AnsiString":
                                return TextEncoding.SafeDecodeString(Encoding.Default, data);
                            default:
                                return data;
                        }
                }
            }
        }

        private RecordDef GetRecordDef(Record record, FormatDef format)
        {
            int index = Type.IndexOf(':') + 1;
            string type_field = Type.Substring(index);
            string type_name = (string)record[type_field];
            RecordDef recordDef = format[type_name];
            return recordDef;
        }

        private int GetSize(Record record)
        {
            int size = Size;
            if (!String.IsNullOrEmpty(SizeField))
            {
                size += Convert.ToInt32(record[SizeField]);
            }
            if (size < 0) size = 0;
            return size;
        }

        public int CalculateSize(Record record, FormatDef format)
        {
            object value = record[Name];
            if (value == null) return 0;
            if (Type.StartsWith("Record"))
            {
                if (value is byte[])
                {
                    return (value as byte[]).Length;
                }
                else
                {
                    //RecordDef recordDef = GetRecordDef(record, format);
                    //return recordDef.CalculateSize(value as Record, format);
                    return (value as Record).Data.Length;
                }
            }
            else if (Type.EndsWith("[]"))
            {
                string type = Type.TrimEnd('[', ']');
                int datasize = 0;
                int size = 0;
                if (type == "Byte")
                {
                    if (value is byte[])
                    {
                        byte[] data = value as byte[];
                        size = data.Length;
                        datasize = size;
                    }
                    else
                    {
                        return 0;
                    }
                }
                else if (value is IEnumerable<Record>)
                {
                    RecordDef recordDef = format[type];
                    IEnumerable<Record> subfields = value as IEnumerable<Record>;
                    foreach (Record subfield in subfields)
                    {
                        if (subfield.HasError)
                        {
                            break;
                        }
                        datasize += recordDef.CalculateSize(subfield, format);
                    }
                    if (value is List<Record>)
                    {
                        size = ((List<Record>)subfields).Count;
                    }
                    else if (value is RecordCollection)
                    {
                        size = ((RecordCollection)subfields).Count;
                    }
                }
                SetSize(record, size - Size);
                return datasize;
            }
            else
            {
                byte[] data = null;
                if (value is byte[])
                {
                    data = value as byte[];
                }
                else
                {
                    switch (Type)
                    {
                        case "Int":
                        case "UInt":
                            return Size;
                        case "AsciiString":
                            data = Encoding.ASCII.GetBytes(value as string);
                            break;
                        case "AnsiString":
                            data = Encoding.Default.GetBytes(value as string);
                            break;
                    }
                }
                SetSize(record, data.Length - Size);
                return data.Length;
            }
        }

        private void SetSize(Record record, int size)
        {
            if (!String.IsNullOrEmpty(SizeField))
            {
                Type type = record[SizeField].GetType();
                record[SizeField] = Convert.ChangeType(size, type);
            }
        }

        public void Encode(DataWriter writer, Record record, FormatDef format)
        {
            object value = record[Name];
            if (Type.StartsWith("Record"))
            {
                if (value is byte[])
                {
                    writer.WriteBytes(value as byte[]);
                }
                else
                {
                    //RecordDef recordDef = GetRecordDef(record, format);
                    //recordDef.Encode(writer, value as Record, format);
                    writer.WriteBytes((value as Record).Data);
                }
            }
            else if (Type.EndsWith("[]"))
            {
                string type = Type.TrimEnd('[', ']');
                if (type == "Byte")
                {
                    if (value is byte[])
                    {
                        writer.WriteBytes(value as byte[]);
                    }
                    else
                    {
                        throw new Exception("Invalid data. " + value);
                    }
                }
                else if (value is List<Record>)
                {
                    RecordDef recordDef = format[type];
                    List<Record> subfields = value as List<Record>;
                    for (int i = 0; i < subfields.Count; i++)
                    {
                        recordDef.Encode(writer, subfields[i], format);
                    }
                }
                else if (value is RecordCollection)
                {
                    RecordDef recordDef = format[type];
                    RecordCollection subfields = value as RecordCollection;
                    subfields.Reset();
                    for (int i = 0; i < subfields.Count; i++)
                    {
                        recordDef.Encode(writer, subfields.Read(), format);
                    }
                }
            }
            else
            {
                string type = Type + (Size * 8) + Endian;
                switch (type)
                {
                    case "UInt8LE":
                    case "UInt8BE":
                        writer.WriteByte((byte)value);
                        break;
                    case "UInt32LE":
                        writer.WriteUInt32LE((UInt32)value);
                        break;
                    case "UInt16BE":
                        writer.WriteUInt16BE((UInt16)value);
                        break;
                    case "UInt32BE":
                        writer.WriteUInt32BE((UInt32)value);
                        break;
                    default:
                        byte[] data = null;
                        if (value is byte[])
                        {
                            data = value as byte[];
                        }
                        else
                        {
                            switch (Type)
                            {
                                case "AsciiString":
                                    data = Encoding.ASCII.GetBytes(value as string);
                                    break;
                                case "AnsiString":
                                    data = Encoding.Default.GetBytes(value as string);
                                    break;
                            }
                        }
                        writer.WriteBytes(data);
                        break;
                }
            }
        }
    }
}
