using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using QiHe.CodeLib;

namespace QiHe.RealMedia
{
    public class PacketRecover
    {
        MediaFile media;
        FormatDef format;
        RecordDef PacketDef;
        Chunk prop = null;
        ushort max_packet_size;
        ushort avg_packet_size;
        uint num_packets = 0;
        uint duration = 0;
        List<ushort> stream_numbers = new List<ushort>();
        List<string> loginfo = new List<string>();

        public List<string> LogInfo
        {
            get { return loginfo; }
        }

        public PacketRecover(MediaFile media, FormatDef format)
        {
            this.media = media;
            this.format = format;
            this.PacketDef = format["MediaPacket"];
            foreach (Chunk chunk in media.Chunks)
            {
                if (chunk.Name == "PROP")
                {
                    prop = chunk;
                    max_packet_size = (ushort)(uint)prop["max_packet_size"];
                    avg_packet_size = (ushort)(uint)prop["avg_packet_size"];
                    duration = (uint)prop["duration"];
                    loginfo.Add(String.Format("duration: {0}", duration));
                    loginfo.Add(String.Format("max_packet_size: {0}", max_packet_size));
                    loginfo.Add(String.Format("avg_packet_size: {0}", avg_packet_size));
                }
                else if (chunk.Name == "MDPR")
                {
                    ushort stream_number = (ushort)chunk["stream_number"];
                    stream_numbers.Add(stream_number);
                }
            }
        }

        public void RecoverPackets()
        {
            Stream stream = new BufferedStream(media.FileStream);
            DataReader reader = new DataReader(stream);

            RecoverIndexChunk(stream, reader);

            Chunk dataChunk = media.GetChunk("DATA");
            stream.Position = dataChunk.Offset + 18;
            uint total_packets = 0;
            num_packets = (uint)dataChunk["num_packets"];
            loginfo.Add(String.Format("num_packets: {0}\r\n", num_packets));

            //List<Record> packets = new List<Record>();
            string tempfile = Path.GetTempFileName();
            RecordCollection packets = new RecordCollection(format, PacketDef, tempfile);

            uint good_offset = 0;
            uint current_offset = 0;
            while (true)
            {
                good_offset = current_offset;
                current_offset = (uint)stream.Position;
                bool isGood;
                Record packet = ReadPacket(reader, out isGood);
                if (isGood)
                {
                    ushort length = (ushort)packet["length"];
                    if (length == (ushort)0)
                    {
                        loginfo.Add(String.Format("Dummy packet: {0}/{1}", packets.Count, num_packets));
                        loginfo.Add(ShowPacket(current_offset, packet));
                    }
                    else
                    {
                        packets.Add(packet);
                    }
                }
                else
                {
                    if (IsINDXorDATA(packet))
                    {
                        break;
                    }
                    loginfo.Add(String.Format("Bad packet found: {0}/{1}", packets.Count, num_packets));
                    loginfo.Add(ShowPacket(current_offset, packet));
                    uint packet_index;
                    uint synchroPoint = FindSynchronizedPoint(good_offset, reader, out packet_index);
                    if (synchroPoint < uint.MaxValue)
                    {
                        //List<uint> offsets = new List<uint>();
                        //offsets.Add(synchroPoint);
                        //List<uint> good_offsets = FindRecoverPoint(offsets, reader, good_offset);
                        //while (good_offsets.Count > 0)
                        //{
                        //    offsets = good_offsets;
                        //    good_offsets = FindRecoverPoint(offsets, reader, good_offset);
                        //}
                        //uint recoverPoint = ChooseRecoverPoint(offsets);
                        uint recoverPoint = synchroPoint;
                        while (synchroPoint < uint.MaxValue)
                        {
                            recoverPoint = synchroPoint;
                            packet_index--;
                            synchroPoint = FindRecoverPoint(recoverPoint, reader, good_offset);
                        }
                        stream.Position = recoverPoint;
                        packet = ReadPacket(reader, out isGood);
                        loginfo.Add(String.Format("Recoverable Point found: {0}/{1}", packet_index, num_packets));
                        loginfo.Add(ShowPacket(recoverPoint, packet));
                        //loginfo.Add(String.Format("Data dropped: {0} - {1} = {2}\r\n", recoverPoint, current_offset, recoverPoint - current_offset));
                        loginfo.Add(String.Format("Search packets between {0} and {1}", current_offset, recoverPoint));
                        List<Record> searchedPackets = SearchPackets(reader, current_offset, recoverPoint);
                        packets.AddRange(searchedPackets);
                        loginfo.Add(String.Format("{0} packets found\r\n", searchedPackets.Count));
                        stream.Position = recoverPoint;
                        current_offset = good_offset;
                    }
                    else
                    {
                        uint end_of_stream = (uint)stream.Length;
                        loginfo.Add(String.Format("Search packets between {0} and {1}", current_offset, end_of_stream));
                        List<Record> searchedPackets = SearchPackets(reader, current_offset, end_of_stream);
                        packets.AddRange(searchedPackets);
                        loginfo.Add(String.Format("{0} packets found\r\n", searchedPackets.Count));
                        break;
                    }
                }
            }
            //packets.Sort(delegate(Record packet1, Record packet2)
            //{
            //    uint timestamp1 = (uint)packet1["timestamp"];
            //    uint timestamp2 = (uint)packet2["timestamp"];
            //    return timestamp1.CompareTo(timestamp2);
            //});
            dataChunk["packets"] = packets;
            dataChunk["num_packets"] = (uint)packets.Count;
            total_packets += (uint)packets.Count;
            loginfo.Add(String.Format("total_packets: {0}", total_packets));

            prop["num_packets"] = total_packets;
        }

        private void RecoverIndexChunk(Stream stream, DataReader reader)
        {
            if (prop != null)
            {
                uint index_offset = (uint)prop["index_offset"];
                Chunk index = media.GetChunk("INDX");
                if (index == null
                    || (index_offset != (uint)index.Offset && index.Record.HasError))
                {
                    while (index_offset > 0)
                    {
                        try
                        {
                            stream.Position = index_offset;
                            index = Reader.DecodeChunk(stream, reader, format);
                            if (index.Name == "INDX")
                            {
                                media.Chunks.Add(index);
                                index_offset = (uint)index["next_index_header"];
                            }
                            else
                            {
                                index_offset = 0;
                            }
                        }
                        catch (EndOfStreamException)
                        {
                            break;
                        }
                    }
                }
            }
        }

        Record ReadPacket(DataReader reader, out bool isGood)
        {
            Record packet = PacketDef.Decode(reader, format);
            if (!packet.HasError && IsGood(packet, stream_numbers))
            {
                isGood = true;
            }
            else
            {
                isGood = false;
            }
            return packet;
        }

        Record ReadPacket(DataReader reader, ushort size, out bool isGood)
        {
            isGood = false;
            Record packet = new Record();
            packet["object_version"] = reader.ReadUInt16BE();
            ushort length = reader.ReadUInt16BE();
            if (length == size)
            {
                packet["length"] = length;
                packet["stream_number"] = reader.ReadUInt16BE();
                packet["timestamp"] = reader.ReadUInt32BE();
                packet["packet_group"] = reader.ReadByte();
                packet["flags"] = reader.ReadByte();
                if (IsGood(packet, stream_numbers))
                {
                    isGood = true;
                }
            }
            return packet;
        }

        Record SearchPacket(DataReader reader, out bool isGood)
        {
            isGood = false;
            Record packet = new Record();
            packet["object_version"] = reader.ReadUInt16BE();
            ushort length = reader.ReadUInt16BE();
            if (length > 12 && length <= max_packet_size)
            {
                packet["length"] = length;
                packet["stream_number"] = reader.ReadUInt16BE();
                uint timestamp = reader.ReadUInt32BE();
                if (timestamp <= duration)
                {
                    packet["timestamp"] = timestamp;
                    packet["packet_group"] = reader.ReadByte();
                    byte flags = reader.ReadByte();
                    if (flags <= 3)
                    {
                        packet["flags"] = flags;
                        if (IsGood(packet, stream_numbers))
                        {
                            packet["data"] = reader.ReadBytes(length - 12);
                            isGood = true;
                        }
                    }
                }
            }
            return packet;
        }

        private static bool IsGood(Record packet, List<ushort> stream_numbers)
        {
            ushort object_version = (ushort)packet["object_version"];
            ushort stream_number = (ushort)packet["stream_number"];
            if (stream_numbers.Contains(stream_number))
            {
                if (object_version == 1)
                {
                    return false;
                    //throw new Exception("packet of version 1 is not supported.");
                }
                else
                {
                    return object_version == 0;
                }
            }
            else
            {
                return false;
            }
        }

        static bool IsINDXorDATA(Record packet)
        {
            ushort object_version = (ushort)packet["object_version"];
            ushort length = (ushort)packet["length"];
            return (object_version == 18766 && length == 17496) // INDX
                || (object_version == 17473 && length == 21569);// DATA

        }

        uint FindSynchronizedPoint(uint good_offset, DataReader reader, out uint packet_index)
        {
            List<uint> offsets = new List<uint>();
            Dictionary<uint, uint> packet_indice = new Dictionary<uint, uint>();
            Dictionary<uint, Record> packets = new Dictionary<uint, Record>();
            foreach (Chunk chunk in media.GetChunks("INDX"))
            {
                IEnumerable<Record> indice = chunk["indices"] as IEnumerable<Record>;
                if (indice != null)
                {
                    foreach (Record indexRecord in indice)
                    {
                        if (indexRecord["offset"] != null && indexRecord["packet_index"] != null)
                        {
                            uint offset = (uint)indexRecord["offset"];
                            if (offset > good_offset)
                            {
                                uint index = (uint)indexRecord["packet_index"];
                                reader.BaseStream.Position = offset;
                                bool isGood;
                                Record packet = ReadPacket(reader, out isGood);
                                if (isGood)
                                {
                                    offsets.Add(offset);
                                    packet_indice.Add(offset, index);
                                    packets.Add(offset, packet);
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            if (offsets.Count > 0)
            {
                uint minOffset = Algorithm.Minimum(offsets);
                packet_index = packet_indice[minOffset];
                loginfo.Add(String.Format("Synchronized Point found: {0}/{1}", packet_index, num_packets));
                loginfo.Add(ShowPacket(minOffset, packets[minOffset]));
                return minOffset;
            }
            else
            {
                packet_index = 0;
                return uint.MaxValue;
            }
        }

        uint FindRecoverPoint(uint synchroPoint, DataReader reader, uint good_offset)
        {
            for (ushort big_size = avg_packet_size = 1; big_size < max_packet_size + 12; big_size++)
            {
                uint try_offset = synchroPoint - big_size;
                if (try_offset > good_offset)
                {
                    if (FindPacket(reader, big_size, try_offset))
                    {
                        return try_offset;
                    }
                }
            }
            for (ushort small_size = avg_packet_size; small_size > 12; small_size--)
            {
                uint try_offset = synchroPoint - small_size;
                if (try_offset > good_offset)
                {
                    if (FindPacket(reader, small_size, try_offset))
                    {
                        return try_offset;
                    }
                }
            }
            return uint.MaxValue;
        }

        private bool FindPacket(DataReader reader, ushort size, uint try_offset)
        {
            reader.BaseStream.Position = try_offset;
            bool isGood;
            Record packet = ReadPacket(reader, size, out isGood);
            return isGood;
        }

        List<Record> SearchPackets(DataReader reader, uint range_start, uint range_end)
        {
            List<Record> packets = new List<Record>();
            try
            {
                reader.BaseStream.Position = range_start;
                for (uint offset = range_start; offset < range_end; )
                {
                    bool isGood;
                    Record packet = SearchPacket(reader, out isGood);
                    if (isGood)
                    {
                        ushort length = (ushort)packet["length"];
                        //loginfo.Add(String.Format("found packet at {0} + {1} = {2}", offset, length, offset + length));
                        //loginfo.Add(ShowPacket(offset, packet));
                        packets.Add(packet);
                        offset += length;
                    }
                    else
                    {
                        offset++;
                        reader.BaseStream.Position = offset;
                    }
                }
            }
            catch (Exception error)
            {
                loginfo.Add(error.Message);
            }
            return packets;
        }

        List<uint> FindRecoverPoint(List<uint> offsets, DataReader reader, uint good_offset)
        {
            List<uint> good_offsets = new List<uint>();
            foreach (uint offset in offsets)
            {
                for (uint size = 1; size <= max_packet_size; size++)
                {
                    uint try_offset = offset - size - 12;
                    if (try_offset > good_offset)
                    {
                        reader.BaseStream.Position = try_offset;
                        bool isGood;
                        Record packet = ReadPacket(reader, out isGood);
                        if (isGood)
                        {
                            loginfo.Add("Recoverable Point found:");
                            loginfo.Add(ShowPacket(try_offset, packet));
                            good_offsets.Add(try_offset);
                        }
                    }
                }
            }
            return good_offsets;
        }

        private uint ChooseRecoverPoint(List<uint> offsets)
        {
            List<int> diff = new List<int>();
            foreach (uint offset in offsets)
            {
                diff.Add((int)Math.Abs(offset - avg_packet_size));
            }
            int minIndex = Algorithm.PosofMinimum(diff);
            return offsets[minIndex];
        }

        static string ShowPacket(uint offset, Record packet)
        {
            StringBuilder text = new StringBuilder();
            text.AppendFormat("         offset : {0}\r\n", offset);
            foreach (KeyValuePair<string, object> field in packet.Fields)
            {
                text.AppendFormat("{0,15} : {1}\r\n", field.Key, field.Value);
            }
            return text.ToString();
        }
    }
}
